package pkg

import "github.com/anchore/syft/internal/evidence"

const (
	EvidenceAnnotationKey        = evidence.AnnotationKey
	PrimaryEvidenceAnnotation    = evidence.PrimaryAnnotation
	SupportingEvidenceAnnotation = evidence.SupportingAnnotation
)
